@@ -14,8 +14,8 @@ namespace cv {
 
 Mat _InputArray::getMat_(int i) const
 {
-    int k = kind();
-    int accessFlags = flags & ACCESS_MASK;
+    _InputArray::KindFlag k = kind();
+    AccessFlag accessFlags = flags & ACCESS_MASK;
 
     if( k == MAT )
     {
@@ -33,12 +33,6 @@ Mat _InputArray::getMat_(int i) const
         return m->getMat(accessFlags).row(i);
     }
 
-    if( k == EXPR )
-    {
-        CV_Assert( i < 0 );
-        return (Mat)*((const MatExpr*)obj);
-    }
-
     if( k == MATX || k == STD_ARRAY )
     {
         CV_Assert( i < 0 );
@@ -132,8 +126,8 @@ Mat _InputArray::getMat_(int i) const
 
 UMat _InputArray::getUMat(int i) const
 {
-    int k = kind();
-    int accessFlags = flags & ACCESS_MASK;
+    _InputArray::KindFlag k = kind();
+    AccessFlag accessFlags = flags & ACCESS_MASK;
 
     if( k == UMAT )
     {
@@ -164,8 +158,8 @@ UMat _InputArray::getUMat(int i) const
 
 void _InputArray::getMatVector(std::vector<Mat>& mv) const
 {
-    int k = kind();
-    int accessFlags = flags & ACCESS_MASK;
+    _InputArray::KindFlag k = kind();
+    AccessFlag accessFlags = flags & ACCESS_MASK;
 
     if( k == MAT )
     {
@@ -179,17 +173,6 @@ void _InputArray::getMatVector(std::vector<Mat>& mv) const
         return;
     }
 
-    if( k == EXPR )
-    {
-        Mat m = *(const MatExpr*)obj;
-        int n = m.size[0];
-        mv.resize(n);
-
-        for( int i = 0; i < n; i++ )
-            mv[i] = m.row(i);
-        return;
-    }
-
     if( k == MATX || k == STD_ARRAY )
     {
         size_t n = sz.height, esz = CV_ELEM_SIZE(flags);
@@ -272,8 +255,8 @@ void _InputArray::getMatVector(std::vector<Mat>& mv) const
 
 void _InputArray::getUMatVector(std::vector<UMat>& umv) const
 {
-    int k = kind();
-    int accessFlags = flags & ACCESS_MASK;
+    _InputArray::KindFlag k = kind();
+    AccessFlag accessFlags = flags & ACCESS_MASK;
 
     if( k == NONE )
     {
@@ -334,7 +317,7 @@ void _InputArray::getUMatVector(std::vector<UMat>& umv) const
 
 cuda::GpuMat _InputArray::getGpuMat() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if (k == CUDA_GPU_MAT)
     {
@@ -360,7 +343,7 @@ cuda::GpuMat _InputArray::getGpuMat() const
 }
 void _InputArray::getGpuMatVector(std::vector<cuda::GpuMat>& gpumv) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if (k == STD_VECTOR_CUDA_GPU_MAT)
     {
         gpumv = *(std::vector<cuda::GpuMat>*)obj;
@@ -368,7 +351,7 @@ void _InputArray::getGpuMatVector(std::vector<cuda::GpuMat>& gpumv) const
 }
 ogl::Buffer _InputArray::getOGlBuffer() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     CV_Assert(k == OPENGL_BUFFER);
 
@@ -376,9 +359,11 @@ ogl::Buffer _InputArray::getOGlBuffer() const
     return *gl_buf;
 }
 
-int _InputArray::kind() const
+_InputArray::KindFlag _InputArray::kind() const
 {
-    return flags & KIND_MASK;
+    KindFlag k = flags & KIND_MASK;
+    CV_DbgAssert(k != EXPR);
+    return k;
 }
 
 int _InputArray::rows(int i) const
@@ -393,7 +378,7 @@ int _InputArray::cols(int i) const
 
 Size _InputArray::size(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -401,12 +386,6 @@ Size _InputArray::size(int i) const
         return ((const Mat*)obj)->size();
     }
 
-    if( k == EXPR )
-    {
-        CV_Assert( i < 0 );
-        return ((const MatExpr*)obj)->size();
-    }
-
     if( k == UMAT )
     {
         CV_Assert( i < 0 );
@@ -515,7 +494,8 @@ Size _InputArray::size(int i) const
 
 int _InputArray::sizend(int* arrsz, int i) const
 {
-    int j, d=0, k = kind();
+    int j, d = 0;
+    _InputArray::KindFlag k = kind();
 
     if( k == NONE )
         ;
@@ -569,7 +549,7 @@ int _InputArray::sizend(int* arrsz, int i) const
     }
     else
     {
-        CV_CheckLE(dims(i), 2, "Not supported");  // TODO Support EXPR with 3+ dims
+        CV_CheckLE(dims(i), 2, "Not supported");
         Size sz2d = size(i);
         d = 2;
         if(arrsz)
@@ -584,7 +564,7 @@ int _InputArray::sizend(int* arrsz, int i) const
 
 bool _InputArray::sameSize(const _InputArray& arr) const
 {
-    int k1 = kind(), k2 = arr.kind();
+    _InputArray::KindFlag k1 = kind(), k2 = arr.kind();
     Size sz1;
 
     if( k1 == MAT )
@@ -618,7 +598,7 @@ bool _InputArray::sameSize(const _InputArray& arr) const
 
 int _InputArray::dims(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -626,12 +606,6 @@ int _InputArray::dims(int i) const
         return ((const Mat*)obj)->dims;
     }
 
-    if( k == EXPR )
-    {
-        CV_Assert( i < 0 );
-        return ((const MatExpr*)obj)->a.dims;
-    }
-
     if( k == UMAT )
     {
         CV_Assert( i < 0 );
@@ -715,7 +689,7 @@ int _InputArray::dims(int i) const
 
 size_t _InputArray::total(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -764,7 +738,7 @@ size_t _InputArray::total(int i) const
 
 int _InputArray::type(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
         return ((const Mat*)obj)->type();
@@ -772,9 +746,6 @@ int _InputArray::type(int i) const
     if( k == UMAT )
         return ((const UMat*)obj)->type();
 
-    if( k == EXPR )
-        return ((const MatExpr*)obj)->type();
-
     if( k == MATX || k == STD_VECTOR || k == STD_ARRAY || k == STD_VECTOR_VECTOR || k == STD_BOOL_VECTOR )
         return CV_MAT_TYPE(flags);
 
@@ -853,7 +824,7 @@ int _InputArray::channels(int i) const
 
 bool _InputArray::empty() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
         return ((const Mat*)obj)->empty();
@@ -861,9 +832,6 @@ bool _InputArray::empty() const
     if( k == UMAT )
         return ((const UMat*)obj)->empty();
 
-    if( k == EXPR )
-        return false;
-
     if( k == MATX || k == STD_ARRAY )
         return false;
 
@@ -925,7 +893,7 @@ bool _InputArray::empty() const
 
 bool _InputArray::isContinuous(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
         return i < 0 ? ((const Mat*)obj)->isContinuous() : true;
@@ -933,7 +901,7 @@ bool _InputArray::isContinuous(int i) const
     if( k == UMAT )
         return i < 0 ? ((const UMat*)obj)->isContinuous() : true;
 
-    if( k == EXPR || k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
+    if( k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
         k == NONE || k == STD_VECTOR_VECTOR || k == STD_BOOL_VECTOR )
         return true;
 
@@ -966,7 +934,7 @@ bool _InputArray::isContinuous(int i) const
 
 bool _InputArray::isSubmatrix(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
         return i < 0 ? ((const Mat*)obj)->isSubmatrix() : false;
@@ -974,7 +942,7 @@ bool _InputArray::isSubmatrix(int i) const
     if( k == UMAT )
         return i < 0 ? ((const UMat*)obj)->isSubmatrix() : false;
 
-    if( k == EXPR || k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
+    if( k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
         k == NONE || k == STD_VECTOR_VECTOR || k == STD_BOOL_VECTOR )
         return false;
 
@@ -1004,7 +972,7 @@ bool _InputArray::isSubmatrix(int i) const
 
 size_t _InputArray::offset(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -1019,7 +987,7 @@ size_t _InputArray::offset(int i) const
         return ((const UMat*)obj)->offset;
     }
 
-    if( k == EXPR || k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
+    if( k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
         k == NONE || k == STD_VECTOR_VECTOR || k == STD_BOOL_VECTOR )
         return 0;
 
@@ -1068,7 +1036,7 @@ size_t _InputArray::offset(int i) const
 
 size_t _InputArray::step(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -1082,7 +1050,7 @@ size_t _InputArray::step(int i) const
         return ((const UMat*)obj)->step;
     }
 
-    if( k == EXPR || k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
+    if( k == MATX || k == STD_VECTOR || k == STD_ARRAY ||
         k == NONE || k == STD_VECTOR_VECTOR || k == STD_BOOL_VECTOR )
         return 0;
 
@@ -1128,7 +1096,7 @@ size_t _InputArray::step(int i) const
 
 void _InputArray::copyTo(const _OutputArray& arr) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == NONE )
         arr.release();
@@ -1137,14 +1105,6 @@ void _InputArray::copyTo(const _OutputArray& arr) const
         Mat m = getMat();
         m.copyTo(arr);
     }
-    else if( k == EXPR )
-    {
-        const MatExpr& e = *((MatExpr*)obj);
-        if( arr.kind() == MAT )
-            arr.getMatRef() = e;
-        else
-            Mat(e).copyTo(arr);
-    }
     else if( k == UMAT )
         ((UMat*)obj)->copyTo(arr);
 #ifdef HAVE_CUDA
@@ -1157,7 +1117,7 @@ void _InputArray::copyTo(const _OutputArray& arr) const
 
 void _InputArray::copyTo(const _OutputArray& arr, const _InputArray & mask) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == NONE )
         arr.release();
@@ -1186,9 +1146,9 @@ bool _OutputArray::fixedType() const
     return (flags & FIXED_TYPE) == FIXED_TYPE;
 }
 
-void _OutputArray::create(Size _sz, int mtype, int i, bool allowTransposed, int fixedDepthMask) const
+void _OutputArray::create(Size _sz, int mtype, int i, bool allowTransposed, _OutputArray::DepthMask fixedDepthMask) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if( k == MAT && i < 0 && !allowTransposed && fixedDepthMask == 0 )
     {
         CV_Assert(!fixedSize() || ((Mat*)obj)->size.operator()() == _sz);
@@ -1228,9 +1188,9 @@ void _OutputArray::create(Size _sz, int mtype, int i, bool allowTransposed, int
     create(2, sizes, mtype, i, allowTransposed, fixedDepthMask);
 }
 
-void _OutputArray::create(int _rows, int _cols, int mtype, int i, bool allowTransposed, int fixedDepthMask) const
+void _OutputArray::create(int _rows, int _cols, int mtype, int i, bool allowTransposed, _OutputArray::DepthMask fixedDepthMask) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if( k == MAT && i < 0 && !allowTransposed && fixedDepthMask == 0 )
     {
         CV_Assert(!fixedSize() || ((Mat*)obj)->size.operator()() == Size(_cols, _rows));
@@ -1271,7 +1231,7 @@ void _OutputArray::create(int _rows, int _cols, int mtype, int i, bool allowTran
 }
 
 void _OutputArray::create(int d, const int* sizes, int mtype, int i,
-                          bool allowTransposed, int fixedDepthMask) const
+                          bool allowTransposed, _OutputArray::DepthMask fixedDepthMask) const
 {
     int sizebuf[2];
     if(d == 1)
@@ -1281,7 +1241,7 @@ void _OutputArray::create(int d, const int* sizes, int mtype, int i,
         sizebuf[1] = 1;
         sizes = sizebuf;
     }
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     mtype = CV_MAT_TYPE(mtype);
 
     if( k == MAT )
@@ -1658,7 +1618,7 @@ void _OutputArray::release() const
 {
     CV_Assert(!fixedSize());
 
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -1726,7 +1686,7 @@ void _OutputArray::release() const
 
 void _OutputArray::clear() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == MAT )
     {
@@ -1745,7 +1705,7 @@ bool _OutputArray::needed() const
 
 Mat& _OutputArray::getMatRef(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if( i < 0 )
     {
         CV_Assert( k == MAT );
@@ -1770,7 +1730,7 @@ Mat& _OutputArray::getMatRef(int i) const
 
 UMat& _OutputArray::getUMatRef(int i) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if( i < 0 )
     {
         CV_Assert( k == UMAT );
@@ -1787,34 +1747,34 @@ UMat& _OutputArray::getUMatRef(int i) const
 
 cuda::GpuMat& _OutputArray::getGpuMatRef() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     CV_Assert( k == CUDA_GPU_MAT );
     return *(cuda::GpuMat*)obj;
 }
 std::vector<cuda::GpuMat>& _OutputArray::getGpuMatVecRef() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     CV_Assert(k == STD_VECTOR_CUDA_GPU_MAT);
     return *(std::vector<cuda::GpuMat>*)obj;
 }
 
 ogl::Buffer& _OutputArray::getOGlBufferRef() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     CV_Assert( k == OPENGL_BUFFER );
     return *(ogl::Buffer*)obj;
 }
 
 cuda::HostMem& _OutputArray::getHostMemRef() const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     CV_Assert( k == CUDA_HOST_MEM );
     return *(cuda::HostMem*)obj;
 }
 
 void _OutputArray::setTo(const _InputArray& arr, const _InputArray & mask) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
 
     if( k == NONE )
         ;
@@ -1838,7 +1798,7 @@ void _OutputArray::setTo(const _InputArray& arr, const _InputArray & mask) const
 
 void _OutputArray::assign(const UMat& u) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if (k == UMAT)
     {
         *(UMat*)obj = u;
@@ -1860,7 +1820,7 @@ void _OutputArray::assign(const UMat& u) const
 
 void _OutputArray::assign(const Mat& m) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if (k == UMAT)
     {
         m.copyTo(*(UMat*)obj); // TODO check m.getUMat()
@@ -1952,7 +1912,7 @@ void _OutputArray::move(Mat& m) const
 
 void _OutputArray::assign(const std::vector<UMat>& v) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if (k == STD_VECTOR_UMAT)
     {
         std::vector<UMat>& this_v = *(std::vector<UMat>*)obj;
@@ -1990,7 +1950,7 @@ void _OutputArray::assign(const std::vector<UMat>& v) const
 
 void _OutputArray::assign(const std::vector<Mat>& v) const
 {
-    int k = kind();
+    _InputArray::KindFlag k = kind();
     if (k == STD_VECTOR_UMAT)
     {
         std::vector<UMat>& this_v = *(std::vector<UMat>*)obj;
